############################### Blonkers! ##############################
# File: screen.py
#
# Imported by: game.py
#
# Description: Contains the Screen class, which handles the creation and
#   updating of the main display window.
#
#
# (C) Copyright 2011 by Chris Flesner
# This file is released under a GNU All-Permissive License:
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.  This file is offered as-is,
# without any warranty.
########################################################################

import pygame
from pygame.locals import *

## Constants
WINDOW_CAPTION = "Blonkers!"

ORIGINAL_WINDOW_SIZE = (1280, 800)
ASPECT_RATIO = (16, 10)
ORIGINAL_SCALE_AMOUNT = .82
TOP_WIDTH = int(ORIGINAL_WINDOW_SIZE[0] / ORIGINAL_SCALE_AMOUNT)
TOP_HEIGHT = int(ORIGINAL_WINDOW_SIZE[1] / ORIGINAL_SCALE_AMOUNT)

## Classes
class Screen:
    """
    self.start_size = original size of the game screen to be scaled to
        the users monitor
    self.display_w, self.display_h = dimensions of user's monitor
    self.window_size = size of window

    self.window = the resized window that the game is played in
    self.surface = the original size surface that the game is drawn to
    
    self.redraw_rects = the rectangles that will need to be redrawn
        to the screen
    self.scaled_redraw_rects = redraw_rects scaled to the window size

    If I have time, I would like to make the window RESIZABLE, but I found
    myself spending far too much time trying to get that to work right to
    implement it yet.  So this is a scaled back version without all the
    non-functioning RESIZABLE-compatible code
    """
    def __init__(self):
        """
        Creates a either a 1280 x 800 window or smaller one if the
        user's monitor is not big enough.
        """
        self.start_size = ORIGINAL_WINDOW_SIZE
        self.videoInfo = pygame.display.Info()
        self.display_w = self.videoInfo.current_w
        self.display_h = self.videoInfo.current_h
        if self.display_w > TOP_WIDTH and self.display_h > TOP_HEIGHT:
            # display is big enough
            self.window_size = self.start_size
        else:
            # scale the window to fit in user's display
            w = int(self.display_w * ORIGINAL_SCALE_AMOUNT)
            h = int(w / ASPECT_RATIO[0] * ASPECT_RATIO[1])
            self.window_size = (w, h)

        ####
        #self.window_size = (320, 200)
        ####
        self.window = pygame.display.set_mode(self.window_size)
        pygame.display.set_caption(WINDOW_CAPTION)

        self.surface = pygame.Surface(ORIGINAL_WINDOW_SIZE)
        
        self.redraw_rects = []
        self.scaled_redraw_rects = []

    def fill(self, colorname, rect = None):
        """
        Fills the window with a solid color.
        """
        color = pygame.color.Color(colorname)
        if rect == None:
            self.surface.fill(color)
            self.redraw_rects = []
            self.redraw_rects.append(self.surface.get_rect())
            return

        pygame.draw.rect(self.surface, color, rect)
        self.redraw_rects.append(rect)

    def blit(self, source, dest, area = None, special_flags = 0):
        self.surface.blit(source, dest, area, special_flags)
        self.redraw_rects.append(dest)
        
    def update(self, update_all = False):
        """
        If update_all is True, updates the entire screen.

        Otherwise, updates the redraw_rects portion of the screens
        if window is 1280x800.  If window is different size then
        it scales the redraw_rects and updates those portions of the
        screen.
        """
        self.resizeAndBlitToScreen()
        
        if update_all:
            pygame.display.update()
            self.redraw_rects = []
            return

        if self.window_size == self.start_size:
            pygame.display.update(self.redraw_rects)
        else:
            self.scaleRedrawRects()
            pygame.display.update(self.scaled_redraw_rects)

        self.redraw_rects = []
    def resizeAndBlitToScreen(self):
        """
        Resizes a copy of self.surface and blits it to self.window
        """
        if self.window_size == self.start_size:
            self.window.blit(self.surface, self.surface.get_rect())
            return
        
        resized_surface = pygame.transform.scale(self.surface.copy(), \
                                                 self.window_size)
        self.window.blit(resized_surface, resized_surface.get_rect())
        
        
    def scaleRedrawRects(self):
        """
        Scales the rectangles in redraw_rects to scaled_redraw_rects
        for the purposes of only updating the parts of the screen
        that need to be redrawn while the window is scaled.

        NOTE: REDRAW RECTANGLES MUST BE PADDED WITH A COUPLE OF EXTRA
        PIXELS FOR THIS TO WORK SMOOTHLY... I think
        """
        self.scaled_redraw_rects = []
        scale_amount = float(self.window_size[0]) / self.start_size[0]
        for i in xrange(len(self.redraw_rects)):
            # first resize the rectangle
            offset_w = int(self.redraw_rects[i].width * scale_amount - \
                           self.redraw_rects[i].width)
            offset_h = int(self.redraw_rects[i].height * scale_amount - \
                           self.redraw_rects[i].height)
            self.scaled_redraw_rects.append(self.redraw_rects[i].copy())
            self.scaled_redraw_rects[i].inflate_ip(offset_w, offset_h)
            # now move it to the correct position
            new_topleft = (int(self.redraw_rects[i].topleft[0] * scale_amount),
                           int(self.redraw_rects[i].topleft[1] * scale_amount))
            self.scaled_redraw_rects[i].topleft = new_topleft
            



#####################################
##def test_screen():
##    import sys
##    pygame.init()
##    screen = Screen()
##    screen.fill('blue')
##    screen.update(True)
##    rects = []
##    rects.append(Rect(600,430,25,25))
##    rects.append(Rect(729,200,25,25))
##    rects.append(Rect(390, 0, 500, 700))
##    rects.append(Rect(420, 119, 25, 25))
##    screen.fill('black', rects[2])
##    screen.fill('red',rects[0])
##    screen.fill('yellow',rects[1])
##    screen.fill('green', rects[3])
##    screen.redraw_rects = rects[:]
##    screen.update()
##
##    while True:
##        screen.curr_window_size = screen.window.get_size()
##        for event in pygame.event.get():
##            if event.type == QUIT:
##                pygame.quit()
##                sys.exit()
##
##test_screen()
