############################### Blonkers! ##############################
# File: sprites.py
#
# Imported by: game.py
#
# Description: contains all the logic for all the sprites used in the game
# including level borders, the zone marker, blocks and block formations
#
#
# (C) Copyright 2011 by Chris Flesner
# This file is released under a GNU All-Permissive License:
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.  This file is offered as-is,
# without any warranty.
########################################################################

import pygame, random, math
from filenames import *

ABOVE_THE_ZONE_CLICK_AMT = 1
BELOW_THE_ZONE_CLICK_AMT = 4

X = 0
Y = 1

LEVEL_SURFACE_SIZE = (500, 700)
BLOCK_SIZE = (25, 25)
LEVEL_SURFACE_LEFT = 390  # in relation to the screen
LEVEL_SURFACE_TOP = 0     #

level_grid = {}
LEVEL_X_GRID = []
assert LEVEL_SURFACE_SIZE[X]%BLOCK_SIZE[X] == 0, \
       'level width % block width must be 0'
for x in xrange(0, LEVEL_SURFACE_SIZE[X], BLOCK_SIZE[X]):
    LEVEL_X_GRID.append(x)

LEVEL_Y_GRID = []
assert LEVEL_SURFACE_SIZE[Y]%BLOCK_SIZE[Y] == 0, \
       'level height % block height must be 0'
for y in xrange(0, LEVEL_SURFACE_SIZE[Y], BLOCK_SIZE[Y]):
    LEVEL_Y_GRID.append(y)

for row in LEVEL_Y_GRID:
    for column in LEVEL_X_GRID:
        level_grid[column, row] = False

## Borders to keep blocks in line
LEFT_BORDER_RECT = pygame.Rect(-30, -LEVEL_SURFACE_SIZE[Y],
                                30, LEVEL_SURFACE_SIZE[Y]*2+30)
RIGHT_BORDER_RECT = pygame.Rect(LEVEL_SURFACE_SIZE[X], -LEVEL_SURFACE_SIZE[Y],
                                30, LEVEL_SURFACE_SIZE[Y]*2+30)
BOTTOM_BORDER_RECT =  pygame.Rect(-30, LEVEL_SURFACE_SIZE[Y],
                                  LEVEL_SURFACE_SIZE[X]+60, 30)

## Used for creating random block formations
FORM_LEFT = 0
FORM_RIGHT = 1
FORM_DOWN = 2
FORM_UP = 3

FORM_DIRS = {None: [FORM_LEFT, FORM_RIGHT, FORM_UP],
             FORM_UP: [FORM_UP, FORM_LEFT, FORM_RIGHT],
             FORM_LEFT: [FORM_LEFT, FORM_UP],
             FORM_RIGHT: [FORM_RIGHT, FORM_UP]}

LEFT_AMOUNT = {FORM_UP: 0,
               FORM_DOWN: 0,
               FORM_LEFT: - BLOCK_SIZE[X],
               FORM_RIGHT: BLOCK_SIZE[X]}

TOP_AMOUNT = {FORM_UP: - BLOCK_SIZE[Y],
              FORM_DOWN: BLOCK_SIZE[Y],
              FORM_LEFT: 0,
              FORM_RIGHT: 0}

## Block colors
GRAY = 0
BLUE = 1
GREEN = 2
LBLUE = 3
ORANGE = 4
RED = 5
YELLOW = 6
COLORS = [GRAY, BLUE, GREEN, LBLUE, ORANGE, RED, YELLOW]

## Used for speed information in BlockFormation class
SPEED = 0
TICKS = 1

## Return values for block formation update()s and other related things
SPLIT_UP = 0 # return value
SPLIT_UP_FALL_SPEED = [9,0]

##
ZONE_MARKER_HEIGHT = 3
ZONE_MARKER_FALL_SPEED = [7,0]

###### Globals
clicked_group = None
collision_groups = []
in_the_zone = []
out_of_the_zone = []
group_lists = [collision_groups, in_the_zone, out_of_the_zone]

###### Non-Block classes
#####
class BorderSprite(pygame.sprite.Sprite):
    def __init__(self, image_file, level_rect, groups = []):
        pygame.sprite.Sprite.__init__(self, groups)
        self.rect = level_rect
        #self.image = pygame.image.load(image_file).convert()
        #self.screen_rect = self.image.get_rect()
#####
class Border(pygame.sprite.Group):
    global collision_groups
    def __init__(self, sprites, at_bottom = False):
        pygame.sprite.Group.__init__(self, sprites)
        collision_groups.append(self)
        self.at_bottom = at_bottom
        self.snap_to_grid_x = False
        self.x_speed = [0,0]
        self.y_speed = [0,0]

#####
class ZoneMarker(pygame.sprite.Sprite):
    def __init__(self):
        pygame.sprite.Sprite.__init__(self, [])
        self.rect = None
        self.setRect(None)
        self.fall_speed = ZONE_MARKER_FALL_SPEED

    def setRect(self, clicked_group):
        top = LEVEL_SURFACE_SIZE[Y] - ZONE_MARKER_HEIGHT
        for group in in_the_zone:
            #if group.x_speed[SPEED] == 0 and \
            #        group.y_speed[SPEED] == 0:
            if group.at_bottom == True and  group != clicked_group:
                group_top = group.getTop()
                if group_top < top:
                    top = group_top

        # move it up a bit
        top -= BLOCK_SIZE[Y] * 4
        self.rect = pygame.Rect(0, top, LEVEL_SURFACE_SIZE[X]-1,
                                ZONE_MARKER_HEIGHT)


zone_marker = ZoneMarker()

###### Base Block Formations
#####
class Block(pygame.sprite.Sprite):
    """
    Base class for block sprites.
    self.image = image
    self.rect = rect
    self.connections = the connections between blocks in a formation
    """
    def __init__(self, image_file_list, color, groups = []):
        pygame.sprite.Sprite.__init__(self, groups)
        self.image = pygame.image.load(image_file_list[color]).convert()
        self.rect = self.image.get_rect()
        self.old_rect = self.rect.copy()

    ###
    def update(self, left, top):
        self.rect.left = left
        self.rect.top = top


#####
class SimpleBlock(Block):
    """
    A basic non-mutated block class.
    """
    def __init__(self, color):
        Block.__init__(self, SIMPLE_BLOCK_FILE_LIST, color)

class StubbornBlock(Block):
    """
    A basic non-mutated block class.
    """
    def __init__(self, color):
        Block.__init__(self, STUBBORN_BLOCK_FILE_LIST, color)


#####
class BlockFormation(pygame.sprite.RenderUpdates):
    """
    The base class for all formations of blocks

    x_speed and y_speed are lists of two integers.  The first integer
    is the number of pixels to move in one clock tick.  However, if
    the first integer is 1 or -1, then the second integer is the number
    of frames to wait before moving a pixel.  For example, if
    x_speed is (1, 1) then the formation moves right one pixel every
    clock tick.  Or, if x_speed is (1, 5), then the formation only moves
    right one pixel every 5 clock ticks.  This allows the movement of
    slow blocks while maintaining a decent framerate for the game.

    x_tick and y_tick are used to count the ticks mentioned in the last
    paragraph.
    """
    def __init__(self, sprites, x_speed, y_speed, fall_speed, falling):
        global collision_groups
        pygame.sprite.RenderUpdates.__init__(self, sprites)
        self.x_speed = x_speed
        self.y_speed = y_speed
        self.slow_down_x = False
        self.slow_down_y = False
        self.slow_down_tick = 0
        self.fall_speed = fall_speed
        self.falling = falling
        self.temp_non_fall = -1
        self.x_tick = 0
        self.y_tick = 0
        self.in_the_zone = False
        self.snap_to_grid_x = False
        self.snap_to_grid_x_infected = False
        self.snap_to_grid_y = False
        self.snap_to_grid_y_infected = False
        self.at_bottom = False
        self.click_timer = 0
        self.num_click = 0
        self.mouse_pos = (0,0)
        self.row_split_up = False
        self.rect = pygame.Rect(0,0,0,0)
        self.old_rect = self.rect.copy()
        self.redraw = True
        self.redraw_infected = False
        self.redraw_sprite_rects = []
        self.drag_limit = None

        collision_groups.append(self)
        out_of_the_zone.append(self)


    ###
    def calcDeltas(self):
        delta_x, delta_y = 0, 0
        if self.x_speed[SPEED] > 1 or self.x_speed[SPEED] < -1 or \
                self.x_speed[SPEED] == 0:
            delta_x = self.x_speed[SPEED]
        else:
            if self.x_tick >= self.x_speed[TICKS]:
                delta_x = self.x_speed[SPEED]
                self.x_tick = 0
            else:
                delta_x = 0

        if self.y_speed[SPEED] > 1 or self.y_speed[SPEED] < -1 or \
                self.y_speed[SPEED] == 0:
            delta_y = self.y_speed[SPEED]
        else:
            if self.y_tick >= self.y_speed[TICKS]:
                delta_y = self.y_speed[SPEED]
                self.y_tick = 0
            else:
                delta_y = 0

        self.x_tick += 1
        self.y_tick += 1

        return delta_x, delta_y

    ###
    def moveSprites(self, delta_x, delta_y, update_old_rect = True):
        for sprite in self.sprites():
            if sprite.old_rect.left % BLOCK_SIZE[X] == 0 \
                and sprite.old_rect.top % BLOCK_SIZE[Y] == 0:
                level_grid[sprite.old_rect.left,
                           sprite.old_rect.top] = False
            if update_old_rect:
                sprite.old_rect = sprite.rect.copy()
                self.redraw_sprite_rects.append(sprite.old_rect.copy())
            sprite.update(sprite.rect.left + delta_x,
                          sprite.rect.top + delta_y)
            if sprite.rect.left % BLOCK_SIZE[X] == 0 \
                    and sprite.rect.top % BLOCK_SIZE[Y] == 0:
                level_grid[sprite.rect.left, sprite.rect.top] = sprite

    def slowDown(self, direction):
        if self.slow_down_tick == 1:
            self.slow_down_tick = 0
            if direction == X:
                if self.x_speed[SPEED] > 1:
                    self.x_speed[SPEED] -= 1
                elif self.x_speed[SPEED] < -1:
                    self.x_speed[SPEED] += 1
                elif self.x_speed[SPEED] == 1 or self.x_speed[SPEED] == -1:
                    self.x_speed[TICKS] += 1
            if direction == Y:
                if self.y_speed[SPEED] > 1:
                    self.y_speed[SPEED] -= 1
                elif self.y_speed[SPEED] < -1:
                    self.y_speed[SPEED] += 1
                elif self.y_speed[SPEED] == 1 or self.y_speed[SPEED] == -1:
                    self.y_speed[TICKS] += 1
        self.slow_down_tick += 1

    ###
    def update(self, zone_marker, clicked_group):
        # is there anythnig underneath it
        if self.rect.bottom != LEVEL_SURFACE_SIZE[Y]:
            clear = 0
            for sprite in self.sprites():
                try:
                    if not level_grid[sprite.rect.left,
                                      sprite.rect.top + BLOCK_SIZE[Y]]:
                        clear += 1
                except:
                    pass
            #print 'clear', clear
            if clear == len(self.sprites()):
                self.at_bottom = False
                self.falling = True
                self.fall_speed = ZONE_MARKER_FALL_SPEED[:]

        if clicked_group == self:
            self.at_bottom = False

        self.redraw_sprite_rects = []
        if self.temp_non_fall > -1:
            self.temp_non_fall += 1
            if self.temp_non_fall == 3:
                self.temp_non_fall = -1
        elif self.falling:
            self.y_speed[SPEED] = self.fall_speed[SPEED]
            self.y_speed[TICKS] = self.fall_speed[TICKS]

        if self.slow_down_x and not self.snap_to_grid_x:
            self.slowDown(X)
        if self.slow_down_y and not self.snap_to_grid_y:
            self.slowDown(Y)

        if self.snap_to_grid_x_infected:
            self.snapToGrid(X)
            snap_to_grid_x_infected = False

        if clicked_group == self:
            delta_x, delta_y = pygame.mouse.get_rel()
            #delta_x = self.mouse_pos[X] - self.getLeft()
            #delta_y = self.mouse_pos[Y] - self.getTop()
            #print delta_x, delta_y
            if delta_x >= BLOCK_SIZE[X]: delta_x = BLOCK_SIZE[X]-1
            if delta_y >= BLOCK_SIZE[Y]: delta_y = BLOCK_SIZE[Y]-1
            if delta_x <= -BLOCK_SIZE[X]: delta_x = -BLOCK_SIZE[X]+1
            if delta_y <= -BLOCK_SIZE[Y]: delta_y = -BLOCK_SIZE[Y]+1

        else:
            delta_x, delta_y = self.calcDeltas()

        if delta_x == 0 and delta_y == 0:
            self.redraw = False
        else:
            self.redraw = True

        # if we are snapping to the grid we may need to stop moving
        if self.snap_to_grid_x:
            if (self.sprites()[0].rect.left + self.x_speed[SPEED]) \
                    % BLOCK_SIZE[X] == 0:
                self.x_speed = [0,0]

        # see if we are in the zone
        if pygame.sprite.spritecollide(zone_marker, self, False):
            #if clicked_group == self:
                #print 'hit bar'
            if self in out_of_the_zone:
                in_the_zone.append(self)
                out_of_the_zone.remove(self)

        if clicked_group == self:
            self.drag_limit = zone_marker.rect.bottom
            if self.rect.top < self.drag_limit:
                delta_y = -(self.rect.top - self.drag_limit)

        # try moving the group
        self.moveSprites(delta_x, delta_y)
        self.rect = self.getRect()

        # find out if we hit anything
        all_cols = []
        if self.redraw:
            for group in collision_groups:
                if group is self:
                    pass
                else:
                    cols = pygame.sprite.groupcollide(self,group,False,False)
                    if cols: # we did
                        group.redraw_infected = True
                        self.at_bottom = group.at_bottom
                        if group.snap_to_grid_x:
                            self.snap_to_grid_x_infected = True
                        all_cols.append(cols)
                    delta_x, delta_y = self.analyzeCollisions(all_cols)

        # move back if we did
        if not (delta_x == 0 and delta_y == 0):
            self.moveSprites(delta_x, delta_y)
            self.rect = self.getRect()

        if not self.redraw and not (delta_x == 0 and delta_y == 0):
            self.redraw = True

        if self.redraw_infected:
            self.redraw = True
            self.redraw_infected = False

        # see if we're entering the zone
        if self.getBottom() >= zone_marker.rect.top and not self.in_the_zone:
            self.in_the_zone = True
            self.falling = True
            self.fall_speed = zone_marker.fall_speed[:]
            self.snapToGrid(X)

        if self.at_bottom:
            self.x_speed = [0,0]
            self.y_speed = [0,0]
            self.fall_speed = [0,0]

        if self.redraw:
            self.rect = self.getRect()

    ###
    def analyzeCollisions(self, all_cols):
        """
        Analyzes the collisions the group had after moving and determines
        how much it needs to move back as to not collide.
        """
        delta_x, delta_y = 0,0
        for cols in all_cols:
            for mysprite in cols.keys():
                for sprite in cols[mysprite]:
                    # did the right side of our sprite hit it?
                    if mysprite.old_rect.right <= sprite.rect.left and \
                            mysprite.rect.right > sprite.rect.left:
                        move_x = sprite.rect.left - (mysprite.rect.right)
                        if move_x < delta_x:
                            delta_x = move_x
                            if self.temp_non_fall:
                                self.x_speed[SPEED] = -self.x_speed[SPEED]
                    # did the left side of our sprite hit it?
                    elif mysprite.old_rect.left >= sprite.rect.right and \
                            mysprite.rect.left < sprite.rect.right:
                        move_x = sprite.rect.right  - mysprite.rect.left
                        if move_x > delta_x:
                            delta_x = move_x
                            if self.temp_non_fall:
                                self.x_speed[SPEED] = -self.x_speed[SPEED]
                    else: pass
                    # did the bottom of our sprite hit it?
                    if mysprite.old_rect.bottom <= sprite.rect.top and \
                            mysprite.rect.bottom > sprite.rect.top:
                        move_y = sprite.rect.top - (mysprite.rect.bottom)
                        if move_y < delta_y:
                            delta_y = move_y
                            if self.temp_non_fall:
                                self.y_speed[SPEED] = -self.y_speed[SPEED]
                    # did the top our or sprite hit it?
                    elif mysprite.old_rect.top >= sprite.rect.bottom and \
                            mysprite.rect.top < sprite.rect.bottom:
                        move_y = sprite.rect.bottom - mysprite.rect.top
                        if move_y > delta_y:
                            delta_y = move_y
                            if self.temp_non_fall:
                                self.y_speed[SPEED] = -self.y_speed[SPEED]
                    else: pass

        return delta_x, delta_y

    ###
    def snapToGrid(self, dimensions):
        if type(dimensions) == type(X):
            if dimensions == X:
                self.snap_to_grid_x = True
                offset = self.sprites()[0].rect.left % BLOCK_SIZE[X]
                if offset > BLOCK_SIZE[X]/2:
                    self.x_speed[SPEED] = BLOCK_SIZE[X] - offset
                    self.x_speed[TICKS] = 0
                else:
                    self.x_speed[SPEED] = -offset
                    self.x_speed[TICKS] = 0

            else:
                assert False, 'snapping to Y is not implemented yet'
        elif type(dimensions) == type(X,Y):
            assert False, 'snapping to X,Y is not implemented yet'
        else:
            assert False, 'invalid input to snapToGrid()'


    ###
    def getRect(self):
        top = LEVEL_SURFACE_SIZE[Y]
        bottom = 0
        left = LEVEL_SURFACE_SIZE[X]
        right = 0
        for sprite in self.sprites():
            if sprite.rect.top < top:
                top = sprite.rect.top
            if sprite.rect.bottom > bottom:
                bottom = sprite.rect.bottom
            if sprite.rect.left < left:
                left = sprite.rect.left
            if sprite.rect.right > right:
                right = sprite.rect.right
        return pygame.Rect(left, top, right-left, bottom+1-top)

    ###
    def getTop(self):
        return self.rect.top

    ###
    def getBottom(self):
        return self.rect.bottom

    ###
    def getLeft(self):
        return self.rect.left

    ###
    def getRight(self):
        return self.rect.right


###### Block Formations in Use
class SingleBlock(BlockFormation):
    def __init__(self, sprite, left, top, x_speed, y_speed,
                 fall_speed = [0,0], falling = True):
        BlockFormation.__init__(self, sprite, x_speed, y_speed,
                                fall_speed, falling)
        assert len(self.sprites()) == 1, 'SingleBlock only holds one sprite'
        self.sprite = sprite
        self.sprite.rect.left = left
        self.sprite.rect.top = top

#####
class MultiBlock(BlockFormation):
    def __init__(self, sprites, left, top, x_speed, y_speed,
                 fall_speed = [0,0], falling = True, instructions = []):
        BlockFormation.__init__(self, sprites, x_speed, y_speed,
                                fall_speed, falling)
        formation_instructions = instructions[:]
        if not formation_instructions:
            # Figure out a random pattern of blocks
            build_dir = None
            for u in xrange(len(sprites)):
                prev_build_dir = build_dir
                build_dir = random.choice(FORM_DIRS[prev_build_dir])
                formation_instructions.append(build_dir)

        # build the block pattern
        sprites = self.sprites()
        left_amt, top_amt = left, top
        for i in xrange(len(sprites)):
            left_amt += LEFT_AMOUNT[formation_instructions[i]]
            top_amt += TOP_AMOUNT[formation_instructions[i]]
            sprites[i].rect.left += left_amt
            sprites[i].rect.top += top_amt
            sprites[i].x_speed, sprites[i].y_speed = x_speed[:], y_speed[:]

        self.clicked = 0
        self.split_up = False

    def update(self, zone_marker, clicked_group):
        BlockFormation.update(self, zone_marker, clicked_group)
        if not self.in_the_zone \
                and self.clicked >=  ABOVE_THE_ZONE_CLICK_AMT:
            return SPLIT_UP
        elif self.clicked >= BELOW_THE_ZONE_CLICK_AMT:
            return SPLIT_UP

